/**
 * @(#)MaileonScheduleService.java
 */
package com.maileon.api.mailings.scheduling;

import com.maileon.api.AbstractMaileonService;
import com.maileon.api.MaileonConfiguration;
import com.maileon.api.MaileonException;
import com.maileon.api.ResponseWrapper;

import java.sql.Date;
import java.sql.Time;

/**
 * The <code>MaileonSchedulingService</code> sends operational requests to the <code>ScheduleResource</code>.
 * 
 */
public class MaileonSchedulingService extends AbstractMaileonService {

	public static final String SERVICE = "MAILEON SCHEDULES";
	
        
        /**
         * Constructs a
         * <code>MaileonSchedulingService</code>.
         *
         * @param config Maileon API-Configuration.
         */
	public MaileonSchedulingService(MaileonConfiguration config) {
		super(config, SERVICE);
	}
	
        /**
         * Creates a new scheduling for a given mailing, using the given date.
         * <p><a href="http://dev.maileon.com/api/rest-api-1-0/mailings/mailing-create-schedule/?lang=en">Documentation website</a></p>
         *
         * @param mailingId The ID of the mailing to set the schedule for
         * @param date The {@link Date}
         * @param hours The hours of the {@link Time} object
         * @param minutes The hours of the {@link Time} object
         * @throws MaileonException
         */
	public void setSchedule(long mailingId, String date, String hours, String minutes) throws MaileonException {
		QueryParameters qp = new QueryParameters("date", date);
		qp.add("hours", hours);
		qp.add("minutes", minutes);
		
		put(String.format("mailings/%s/schedule", String.valueOf(mailingId)), qp, null);
	}

	/**
	 * Creates a new scheduling for a given mailing
	 * <p><a href="http://dev.maileon.com/api/rest-api-1-0/mailings/mailing-create-schedule/?lang=en">Documentation website</a></p>
	 *
	 * @param mailingId The ID of the mailing to set the schedule for
	 * @param schedule The {@link Scheduling}
	 * @throws MaileonException
	 */
	public ResponseWrapper setSchedule(long mailingId, Scheduling schedule) throws MaileonException {
		QueryParameters qp = new QueryParameters("date", schedule.getDate());
		qp.add("hours", schedule.getHours());
		qp.add("minutes", schedule.getMinutes());

		return put(String.format("mailings/%s/schedule", String.valueOf(mailingId)), qp, null);
	}

        /**
         * Updates the current scheduling for a given mailing, using the given date.
         * <p><a href="http://dev.maileon.com/api/rest-api-1-0/mailings/mailing-update-schedule/?lang=en">Documentation website</a></p>
         *
         * @param mailingId The ID of the mailing to update the schedule for
         * @param date The {@link Date}
         * @param hours The hours of the {@link Time} object
         * @param minutes The hours of the {@link Time} object
         * @throws MaileonException
         */
	public void updateSchedule(long mailingId, String date, String hours, String minutes) throws MaileonException {
		QueryParameters qp = new QueryParameters("date", date);
		qp.add("hours", hours);
		qp.add("minutes", minutes);
		
		post(String.format("mailings/%s/schedule", String.valueOf(mailingId)), qp, null);
	}
	
	// TODO: returns? Create wrapper datatype
	/**
	 * Gets the details of the schedule for a given mailing.
         * <p><a href="http://dev.maileon.com/api/rest-api-1-0/mailings/mailing-get-schedule/?lang=en">Documentation website</a></p>
         * 
	 * @param mailingId he ID of the mailing to get the schedule for 
         * @return The corresponding {@link Scheduling}
	 * @throws MaileonException
	 */
	public Scheduling getSchedule(long mailingId) throws MaileonException {
		ResponseWrapper response = get(String.format("mailings/%s/schedule", String.valueOf(mailingId)), null);
		return new Scheduling(
				response.getEntityAsXml().selectSingleNode("./date").getText(),
				response.getEntityAsXml().selectSingleNode("./hours").getText(),
				response.getEntityAsXml().selectSingleNode("./minutes").getText(),
				response.getEntityAsXml().selectSingleNode("./state").getText());
	}
	
        /**
         * Deletes the scedule for a given mailing.
         * <p><a
         * href="http://dev.maileon.com/api/rest-api-1-0/mailings/mailing-delete-schedule/?lang=en">Documentation website</a></p>
         *
         * @param mailingId The ID of the mailing to delete the schedule for
         * @throws MaileonException
         */
	public void deleteSchedule(long mailingId) throws MaileonException {
		delete(String.format("mailings/%s/schedule", String.valueOf(mailingId)), null);
	}
}
