package com.maileon.api.transactions;

import com.maileon.api.*;
import com.maileon.api.transactions.serializer.TransactionProcessingReportsJsonSerializer;
import com.maileon.api.transactions.serializer.TransactionTypeXmlSerializer;
import com.maileon.api.transactions.serializer.TransactionTypesXmlSerializer;
import com.maileon.api.transactions.serializer.TransactionsJsonSerializer;
import com.maileon.api.utils.PageUtils;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.MediaType;
import java.util.Arrays;
import java.util.List;

/**
 * The <code>MaileonTransactionsService</code> client send operational requests to the <code>TransactionsResource</code>.
 *
 */
public class MaileonTransactionsService extends AbstractMaileonService {

    public static final String SERVICE = "MAILEON TRANSACTIONS SERVICE";

    /**
     * Instantiates a new maileon transaction service.
     *
     * @param config the configuration of maileon API
     */
    public MaileonTransactionsService(MaileonConfiguration config) {
        super(config, SERVICE);
    }

    /**
     * Returns the overall number of transaction types in the account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/get-transaction-types-count/?lang=en">Documentation website</a></p>
     *
     * @return the count of transaction types
     * @throws MaileonException
     */
    public int getTransactionTypesCount() throws MaileonException {
        ResponseWrapper response = get("transactions/types/count");
        return Integer.parseInt(response.getEntityAsXml().getText());
    }

    /**
     * Returns a page (list) of all transaction types defined in the account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/get-all-transaction-types/?lang=en">Documentation website</a></p>
     *
     * @param pageIndex strating from 1
     * @param pageSize the number transaction types of within a page
     * @return the page containing the found transaction types as {@link TransactionType} objects
     * @throws MaileonException
     */
    public Page<TransactionType> getTransactionTypes(int pageIndex, int pageSize) throws MaileonException {
        QueryParameters params = new QueryParameters("page_index", pageIndex);
        params.add("page_size", pageSize);
        ResponseWrapper response = get("transactions/types", params);
        Page<TransactionType> page = PageUtils.createPage(pageIndex, pageSize, response);
        page.setItems(TransactionTypesXmlSerializer.deserialize(response.getEntityAsXml()));
        return page;
    }

    /**
     * Creates a new transaction type with the given name in the account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/create-transaction-type/?lang=en">Documentation website</a></p>
     *
     * @param type the TransactionType to create
     * @throws MaileonException
     */
    public long createTransactionType(TransactionType type) throws MaileonException {
        ResponseWrapper response = post("transactions/types", TransactionTypeXmlSerializer.serialize(type).asXML());
        if (response.getStatusCode() != 200) {
            throw new WebApplicationException(response.getStatusCode());
        }
        return Long.parseLong(response.getEntityAsXml().getText());
    }

    /**
     * Delete a transaction type from the account according to a given name.
     *
     * @param typeName the name of transaction type
     * @throws MaileonException
     */
    public void deleteTransactionTypeByName(String typeName) throws MaileonException {
        TransactionType type = findTransactionTypeByName(typeName);
        if (type != null) {
            deleteTransactionType(type.getId());
        }
    }

    /**
     * Searches for a patricular transaction type in the account according to a given name.
     *
     * @param typeName the name of transaction type to find
     * @return the found TransactionType correspondingly
     * @throws MaileonException
     */
    public TransactionType findTransactionTypeByName(String typeName) throws MaileonException {
        int numPages = 2;
        for (int i = 1; i <= numPages; i++) {
            Page<TransactionType> page = getTransactionTypes(i, 100);
            numPages = page.getCountPages();
            for (TransactionType type : page.getItems()) {
                if (type.getName().equals(typeName)) {
                    return type;
                }
            }
        }
        return null;
    }

    /**
     * Delete a transaction type from the account according to a given id.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/delete-transaction-type/?lang=en">Documentation website</a></p>
     *
     * @param id the id of transaction type to delete
     * @throws MaileonException
     */
    public void deleteTransactionType(long id) throws MaileonException {
        ResponseWrapper response = delete("transactions/types/" + id);
        if (response.getStatusCode() < 200 || response.getStatusCode() >= 300) {
            throw new WebApplicationException(response.getStatusCode());
        }
    }

    /**
     * Retrieves the transaction type with the given id in the account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/delete-transaction-type/?lang=en">Documentation website</a></p>
     *
     * @param id the id of the required transaction type
     * @return the corresponding TransactionType
     * @throws MaileonException
     */
    public TransactionType getTransactionType(long id) throws MaileonException {
        ResponseWrapper response = get("transactions/types/" + Long.toString(id));
        return TransactionTypeXmlSerializer.deserialize(response.getEntityAsXml());
    }

    /**
     * Creates a single transation in the account.
     *
     * @param transaction the Transaction to create
     * @param ignoreInvalidTransactions if set to <code>false</code>, exceptions like invalid contacts will cause the service to return <code>400 Bad request</code>
     * @return a report of type {@link TransactionProcessingReport} hat indicates if the queuing was successful for the underlying transaction
     * @throws MaileonException
     */
    public TransactionProcessingReport createTransaction(Transaction transaction, boolean ignoreInvalidTransactions) throws MaileonException {
        return createTransactions(Arrays.asList(transaction), ignoreInvalidTransactions).iterator().next();
    }

    /**
     * Creates a single transation in the account.
     *
     * @param transaction the Transaction to create
     * @param release parameter is ignored
     * @param ignoreInvalidTransactions if set to <code>false</code>, exceptions like invalid contacts will cause the service to return <code>400 Bad request</code>
     * @return a report of type {@link TransactionProcessingReport} hat indicates if the queuing was successful for the underlying transaction
     * @throws MaileonException
     * @deprecated use the version without release-parameter
     */
    public TransactionProcessingReport createTransaction(Transaction transaction, boolean release, boolean ignoreInvalidTransactions) throws MaileonException {
        return createTransactions(Arrays.asList(transaction), ignoreInvalidTransactions).iterator().next();
    }

    /**
     * Creates any number of transactions in the account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/create-transactions/?lang=en">Documentation website</a></p>
     *
     * @param transactions the list of {@link Transaction} to create
     * @param ignoreInvalidTransactions if set to <code>false</code>, exceptions like invalid contacts will cause the service to return <code>400 Bad request</code>
     * @return a list of reports of type {@link TransactionProcessingReport} that indicates if the queuing was successful for each transaction
     * @throws MaileonException
     */
    public List<TransactionProcessingReport> createTransactions(List<Transaction> transactions, boolean ignoreInvalidTransactions) throws MaileonException {
        QueryParameters params = new QueryParameters("ignore_invalid_transactions", ignoreInvalidTransactions);
        ResponseWrapper response = post("transactions", params, MediaType.APPLICATION_JSON_TYPE, TransactionsJsonSerializer.serialize(transactions).toJSONString());
        if (response.hasEntity()) {
            return TransactionProcessingReportsJsonSerializer.deserialize(response.getEntityAsString());
        } else {
            throw new WebApplicationException(response.getStatusCode());
        }
    }

    /**
     * Creates any number of transactions in the account.
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/create-transactions/?lang=en">Documentation website</a></p>
     *
     * @param transactions the list of {@link Transaction} to create
     * @param release parameter is ignored
     * @param ignoreInvalidTransactions if set to <code>false</code>, exceptions like invalid contacts will cause the service to return <code>400 Bad request</code>
     * @return a list of reports of type {@link TransactionProcessingReport} that indicates if the queuing was successful for each transaction
     * @throws MaileonException
     * @deprecated use the version without release-parameter
     */
    public List<TransactionProcessingReport> createTransactions(List<Transaction> transactions, boolean release, boolean ignoreInvalidTransactions) throws MaileonException {
        return createTransactions(transactions, ignoreInvalidTransactions);
    }

    /**
     * <p>
     * Delete all transactions of a given type before a given date in the account. Any previously-released transactions will be ignored.</p>
     * <p>
     * <a href="http://dev.maileon.com/api/rest-api-1-0/transactions/delete-transactions-before-a-given-date/?lang=en">Documentation website</a></p>
     *
     * @param typeId the transaction type id of the transactions to delete
     * @param beforeTimestamp the timestamp to compare against, in milliseconds
     * @throws MaileonException
     */
    public void deleteTransactions(long typeId, long beforeTimestamp) throws MaileonException {
        QueryParameters params = new QueryParameters("type_id", typeId);
        params.add("before_timestamp", beforeTimestamp);
        delete("transactions", params);
    }

}
