package com.maileon.api.contactfilters;

import com.maileon.api.MaileonException;
import com.maileon.api.Page;
import com.maileon.api.test.TestProperties;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Test driver for the contact filter service
 *
 * @author Felix Heinrichs | Trusted Mails GmbH | <a href="mailto:felix.heinrichs@trusted-mails.com">felix.heinrichs@trusted-mails.com</a>
 * @author Marcus St&auml;nder | Trusted Mails GmbH | <a href="mailto:marcus.staender@trusted-mails.com">marcus.staender@trusted-mails.com</a>
 *
 */
public class MaileonContactFiltersTestDriver {

    private static final long DEFAULT_CONTACT_FILTER_ID = 2765l;
    private static final long DEFAULT_CONTACT_FILTER_ID_NOT_FOUND = 999l;

    private static TestProperties testProperties;
    protected static MaileonContactFiltersService service = null;



    @BeforeClass
    public static void setUp() throws MaileonException {
        testProperties = new TestProperties();
        service = new MaileonContactFiltersService(testProperties.getMaileonConfiguration());
        service.setDebug(testProperties.isDebug());
    }

    /**
     * Get contact filters count
     */
    @Test
    public void testRetrieveContactFiltersCount() throws Throwable {
        String msg = " - Testing 'retrieve contact filters count'";
        try {
            int result = service.getContactFiltersCount();
            TestProperties.report(msg, true, Integer.toString(result), null);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Get all contact filters
     */
    @Test
    public void testRetrieveContactFilters() throws Throwable {
        String msg = " - Testing 'retrieve all contact filters'";
        try {
            Page<ContactFilter> filters = service.getContactFilters(4, 1); // PageSize: 20, pageIndex 1
            if (filters != null) {
                StringBuilder sb = new StringBuilder();
                for (ContactFilter filter:filters.getItems()) {
                    sb.append(filter.toString()).append("\n");
                }
                TestProperties.report(msg, true, sb.toString(), null);
            } else {
                TestProperties.report(msg, true, "No filters returned (null)", null);
            }
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Get a single contact filter
     */
    @Test
    public void testRetrieveContactFilter() throws Throwable {
        String msg = " - Testing 'retrieve a single contact filter'";
        try {
            Page<ContactFilter> filters = service.getContactFilters(1, 1);
            ContactFilter filter = service.getContactFilter(filters.getItems().get(0).getId());
            TestProperties.report(msg, true, filter.toString(), null);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Get a single contact filter (fails, ID not found)
     */
    @Test (expected = Exception.class)
    public void testRetrieveContactFilterNotFound() throws Throwable {
        String msg = " - Testing 'retrieve a single contact filter' that fails (404)";
        try {
            ContactFilter filter = service.getContactFilter(DEFAULT_CONTACT_FILTER_ID_NOT_FOUND);
            TestProperties.report(msg, true, filter.toString(), null);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            Assert.assertEquals(404, e.getStatusCode());
            throw e;
        }
    }

    /**
     * Update a single contact filter
     */
    @Test
    public void testUpdateContactFilter() throws Throwable {
        String msg = " - Testing 'updating a single contact filter'";
        try {Page<ContactFilter> filters = service.getContactFilters(1, 1);
            service.updateContactFilter(filters.getItems().get(0).getId(), new ContactFilter(1, "newFilterName", "2013-05-21 12:30:23", null, 0, 0, null, 0, filters.getItems().get(0).getRules()));
            TestProperties.report(msg, true, null, null);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Refresh a single contact filter
     */
    @Test
    public void testRefreshContactFilterContacts() throws Throwable {
        String msg = " - Testing 'refreshing a single contact filter'";
        try {
            Page<ContactFilter> filters = service.getContactFilters(1, 1);
            service.refreshContactFilterContacts(filters.getItems().get(0).getId(), Long.MAX_VALUE);
            TestProperties.report(msg, true, null, null);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }
}
