package com.maileon.api.contacts;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import com.maileon.api.MaileonBadRequestException;
import com.maileon.api.MaileonException;
import com.maileon.api.Page;
import com.maileon.api.contactfilters.ContactFilter;
import com.maileon.api.contactfilters.MaileonContactFiltersService;
import com.maileon.api.test.TestProperties;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.junit.BeforeClass;
import org.junit.Test;

import java.io.ByteArrayInputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.UUID;

/**
 * Test driver for the contact service
 *
 * @author Felix Heinrichs | Trusted Mails GmbH | <a href="mailto:felix.heinrichs@trusted-mails.com">felix.heinrichs@trusted-mails.com</a>
 * @author Marcus St&auml;nder | Trusted Mails GmbH | <a href="mailto:marcus.staender@trusted-mails.com">marcus.staender@trusted-mails.com</a>
 *
 */
public class MaileonContactTestDriver {

    private static final int DEFAULT_CONTACT_FILTER_ID = 2765;
    private static final int DEFAULT_CONTACT_ID = 1;
    private static final String DEFAULT_EMAIL = "max.mustermann@xqueue.de";
    private static final String DEFAULT_EXTERNAL_ID = "1337";

    private static TestProperties testProperties;
    protected static MaileonContactsService service = null;

    @BeforeClass
    public static void setUp() throws MaileonException {
        testProperties = new TestProperties();
        service = new MaileonContactsService(testProperties.getMaileonConfiguration());
        service.setDebug(testProperties.isDebug());
    }

    @Test
    public void testDeserializeContact() throws Throwable {
        String contactString = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                + "<contact>"
                + "<email>test@test.com</email>"
                + "<standard_fields>"
                + "<field><name><![CDATA[SALUTATION]]></name><value><![CDATA[Herr]]></value></field>"
                + "<field><name><![CDATA[TITLE]]></name><value><![CDATA[DI]]></value></field>"
                + "<field><name><![CDATA[FIRSTNAME]]></name><value><![CDATA[Thomas]]></value></field>"
                + "<field><name><![CDATA[LASTNAME]]></name><value><![CDATA[Muster]]></value></field>"
                + "<field><name><![CDATA[COUNTRY]]></name><value><![CDATA[Ã–sterreich]]></value></field>"
                + "<field><name><![CDATA[LOCALE]]></name><value><![CDATA[de_DE]]></value></field>"
                + "</standard_fields>"
                + "<custom_fields>"
                + "<field><name><![CDATA[Unternehmen]]></name><value><![CDATA[elements.at]]></value></field>"
                + "<field><name><![CDATA[Position]]></name><value><![CDATA[CEO]]></value></field>"
                + "<field><name><![CDATA[Sprachauswahl]]></name><value><![CDATA[de_DE]]></value></field>"
                + "</custom_fields>"
                + "</contact>";
        Contact contact = ContactAdaptor.fromXml(loadXMLFromString(contactString));
        assertEquals(contact.getEmail(), "test@test.com");
    }

    private Element loadXMLFromString(String xml) throws Exception {
        SAXReader reader = new SAXReader();
        reader.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
        reader.setFeature("http://xml.org/sax/features/external-general-entities", false);
        reader.setFeature("http://xml.org/sax/features/external-parameter-entities", false);
        Document document = reader.read(new ByteArrayInputStream(xml.getBytes(StandardCharsets.UTF_8)));
        return document.getRootElement();
    }

    @Test
    public void testPizzaContact() throws Throwable {
        Contact contact = new Contact();
        contact.setEmail(DEFAULT_EMAIL);
        contact.setExternalId("1338");
        contact.getStandardFields().put(StandardContactField.FIRSTNAME, "Marcus");

        try {
            service.createContact(contact, SynchronizationMode.UPDATE, null, null, true, true, "");
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Contact created.", null);
            }

            List<StandardContactField> scf = new ArrayList<>();
            scf.add(StandardContactField.FIRSTNAME);
            List<String> ccf = new ArrayList<>();
            service.getContactsByEmail(DEFAULT_EMAIL, scf, ccf);
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed creating contact.", null);
            throw e;
        }
    }

    @Test
    public void testCreateContact() throws Throwable {
        Contact contact = new Contact();
        contact.setEmail(DEFAULT_EMAIL);
        contact.setPermission(Permission.NONE);
        contact.getStandardFields().put(StandardContactField.FIRSTNAME, "MAX");

        try {
            service.createContact(contact, SynchronizationMode.UPDATE, null, null, true, true, "");
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Contact created.", null);
            }
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed creating contact.", null);
            throw e;
        }
    }

    @Test
    public void testCreateContactByExternalId() throws Throwable {
        Contact contact = new Contact();
        contact.setEmail("max.mustermann@xqueue.com");
        contact.setExternalId("1339");
        contact.setPermission(Permission.DOI_PLUS);
        contact.getStandardFields().put(StandardContactField.FIRSTNAME, "Name: 1339");

        try {
            service.createContact(contact, SynchronizationMode.UPDATE, null, null, true, true, "", true);
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Contact created.", null);
            }
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed creating contact.", null);
            throw e;
        }
    }

    @Test(expected = MaileonBadRequestException.class)
    public void testCreateContactByExternalIdWithoutValidEmail() throws Throwable {
        Contact contact = new Contact();
        contact.setEmail("INVALID_EMAIL");
        contact.setExternalId("1339");
        contact.setPermission(Permission.DOI_PLUS);
        contact.getStandardFields().put(StandardContactField.FIRSTNAME, "Name: 1339");

        try {
            service.createContact(contact, SynchronizationMode.UPDATE, null, null, true, true, "", true);
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Contact created.", null);
            }
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed creating contact.", null);
            throw e;
        }
    }

    @Test
    public void testCreateContactByExternalIdLifecycle() throws Throwable {

        // Some operations seem to be not complete even if API call returns
        int delay = 1000;

        Contact c1 = new Contact();
        c1.setEmail("maxmustermann1@xqueue.de");
        c1.setExternalId("ExtId1");
        c1.setPermission(Permission.DOI_PLUS);
        c1.getStandardFields().put(StandardContactField.FIRSTNAME, "Name: C1");

        // Make sure no such contacts exist
        try {
            service.deleteContactsByExternalId(c1.getExternalId());
        } catch (Exception e) {
        }

        // Wait a second
        Thread.sleep(delay);

        // Getting a contact must result in a list with 0 entries
        List<Contact> contacts = service.getContactsByExternalId(c1.getExternalId(), null, null);
        assertEquals(contacts.size(), 0);

        try {
            service.createContact(c1, SynchronizationMode.UPDATE, null, null, false, false, "", true);
        } catch (MaileonException e) {
            fail("Contalct should be created without exception");
        }

        // Wait a second
        Thread.sleep(delay);

        // Getting a contact must result in a contact list of length 1
        contacts = service.getContactsByExternalId(c1.getExternalId(), null, null);
        assertEquals(contacts.size(), 1);

        // Try changing the Emailaddress
        c1.setEmail("maxmustermann1.1@xqueue.de");

        try {
            service.createContact(c1, SynchronizationMode.UPDATE, null, null, false, false, "", true);
        } catch (MaileonException e) {
            fail("Contalct should be updated without exception");
        }

        // Wait a second
        Thread.sleep(delay);

        // Getting a contact must still result in a contact list of length 1
        contacts = service.getContactsByExternalId(c1.getExternalId(), null, null);
        assertEquals(contacts.size(), 1);

        // Now lets duplicate the external ID using the (unique email) to create the contact
        // Make sure no such contacts exist
        c1.setEmail("maxmustermann1@xqueue.de");
        try {
            service.deleteContactsByEmail(c1.getEmail());
        } catch (Exception e) {
        }

        // Wait a second
        Thread.sleep(delay);

        try {
            service.createContact(c1, SynchronizationMode.UPDATE, null, null, false, false, "", false);
        } catch (MaileonException e) {
            fail("Contalct should be updated without exception");
        }

        // Wait a second
        Thread.sleep(delay);

        // Getting a contact must still result in a contact list of length 1
        contacts = service.getContactsByExternalId(c1.getExternalId(), null, null);
        assertEquals(contacts.size(), 2);

        // Now update by external ID should fail
        // Wait a second
        Thread.sleep(delay);

        try {
            service.createContact(c1, SynchronizationMode.UPDATE, null, null, false, false, "", true);
            fail("Contalct should fail updating as not unique");
        } catch (MaileonException e) {
        }

        // Make sure no such contacts exist
        try {
            service.deleteContactsByExternalId(c1.getExternalId());
        } catch (Exception e) {
        }
    }

    @Test
    public void testCreateContactIllegalBirthdate() throws Throwable {
        Contact contact = new Contact();
        contact.setEmail(DEFAULT_EMAIL);
        contact.setExternalId("1337");
        contact.getStandardFields().put(StandardContactField.FIRSTNAME, "");
        contact.getStandardFields().put(StandardContactField.BIRTHDAY, "N/A");

        try {
            service.createContact(contact, SynchronizationMode.UPDATE, null, null, false, false, "");
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Contact created.", null);
            }
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed creating contact.", null);
            throw e;
        }
    }

    @Test
    public void testGetContactsByEmail() throws Throwable {
        List<StandardContactField> standardFields = new ArrayList<>();
        standardFields.add(StandardContactField.FIRSTNAME);
        standardFields.add(StandardContactField.LOCALE);

        List<String> customFields = new ArrayList<>();

        try {
            List<Contact> contacts = service.getContactsByEmail(DEFAULT_EMAIL, standardFields, customFields); //DEFAULT_EMAIL
            for (Contact contact : contacts) {
                System.out.println(contact);
            }
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Success getting contacts by email.", null);
            }
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed getting contacts by email.", null);
            throw e;
        }
    }

    @Test
    public void testGetContactsByExternalId() throws Throwable {
        List<StandardContactField> standardFields = new ArrayList<>();
        standardFields.add(StandardContactField.FIRSTNAME);
        standardFields.add(StandardContactField.LOCALE);

        List<String> customFields = new ArrayList<>();

        try {
            List<Contact> contacts = service.getContactsByExternalId("1338", standardFields, customFields); //DEFAULT_EMAIL
            for (Contact contact : contacts) {
                System.out.println(contact);
            }
            if (testProperties.isReportOnSuccess()) {
                TestProperties.report("", true, "Success getting contacts by email.", null);
            }
        } catch (MaileonException e) {
            TestProperties.report("", false, "Failed getting contacts by email.", null);
            throw e;
        }
    }

    @Test
    public void testUnsubscribeContactsByExternalId() throws Throwable {
        try {
            service.unsubscribeContactByExternalId(DEFAULT_EXTERNAL_ID);
        } catch (MaileonException e) {
            TestProperties.report("Failed unsubscribing contacts by external id.", false, null, e);
            throw e;
        }
    }

    @Test
    public void testUpdateContact() throws Throwable {
        String msg = " - Testing 'updating account 1'";
        try {
            Contact newContact = new Contact();
            newContact.setId(128l);
            newContact.setExternalId("1337");
            newContact.setEmail("hamid.irani@gmx.de");
            newContact.getStandardFields().put(StandardContactField.FIRSTNAME, "ccc");
            //newContact.getCustomFields().put("someKey", "someValue");
            service.updateContact(newContact, "ECVmWeGH_1s", false, null, null, null, true);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testUnsubscribeContactById() throws Throwable {
        String msg = " - Testing 'unsubscribing contact with ID " + DEFAULT_CONTACT_ID + "'";
        try {
            service.unsubscribeContactById(DEFAULT_CONTACT_ID);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testDeleteContactsByMaileonId() throws Throwable {
        String msg = " - Testing 'deleting contact with maileon id " + DEFAULT_CONTACT_ID + "'";
        try {
            service.deleteContactsByMaileonId(DEFAULT_CONTACT_ID);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testUnsubscribeContactsByEmail() throws Throwable {
        String msg = " - Testing 'unsubscribing contact(s) with email " + DEFAULT_EMAIL + "'";
        try {
            service.unsubscribeContactsByEmail(DEFAULT_EMAIL);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testDeleteContactsByEmail() throws Throwable {
        String msg = " - Testing 'deleting contact(s) with email " + DEFAULT_EMAIL + "'";
        try {

            // First make sure the contact exists
            Contact contact = new Contact();
            contact.setEmail(DEFAULT_EMAIL);

            try {
                service.createContact(contact, SynchronizationMode.UPDATE, null, null, false, false, "");
            } catch (MaileonException e) {
                // Ignore, contact probably already existed
            }

            service.deleteContactsByEmail(DEFAULT_EMAIL);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testSynchronizeContacts() throws Throwable {
        String msg = " - Testing 'synchronizing contact 1'";
        try {
            List<Contact> contacts = new ArrayList<>();

            Contact newContact = new Contact();
            newContact.setEmail("maxmustermann@xqueue.de");
            newContact.setExternalId("1");

            newContact.getStandardFields().put(StandardContactField.FIRSTNAME, UUID.randomUUID().toString());
            newContact.getStandardFields().put(StandardContactField.LASTNAME, UUID.randomUUID().toString());
//            newContact.getCustomFields().put("Synchronization_Test_Field", UUID.randomUUID().toString());
            contacts.add(newContact);

            newContact = new Contact();
            newContact.setEmail("maxmustermann1@xqueue.de");
            newContact.setExternalId("2");

            newContact.getStandardFields().put(StandardContactField.FIRSTNAME, UUID.randomUUID().toString());
            newContact.getStandardFields().put(StandardContactField.LASTNAME, UUID.randomUUID().toString());
//            newContact.getCustomFields().put("Synchronization_Test_Field", UUID.randomUUID().toString());
            contacts.add(newContact);

            service.synchronizeContacts(contacts, Permission.DOI_PLUS, SynchronizationMode.UPDATE, false, false, true, false, false);
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testReimportUnsubscribedContactsOnSynchronizeContactsLifecycle() throws Throwable {
        String msg = " - Testing 'testReimportUnsubscribedContactsOnSynchronizeContacts'";
        try {

            // Some operations seem to be not complete even if API call returns
            int delay = 3000;
            int retries = 10;

            List<Contact> contacts = new ArrayList<>();

            Contact c1 = new Contact();
            c1.setEmail("maxmustermann1@xqueue.de");
            c1.getStandardFields().put(StandardContactField.FIRSTNAME, "Max1");
            c1.getStandardFields().put(StandardContactField.LASTNAME, "Mustermann1");
            contacts.add(c1);

            Contact c2 = new Contact();
            c2.setEmail("maxmustermann2@xqueue.de");
            c2.getStandardFields().put(StandardContactField.FIRSTNAME, "Max2");
            c2.getStandardFields().put(StandardContactField.LASTNAME, "Mustermann2");
            contacts.add(c2);

            // First make sure both are deleted
            try {
                service.deleteContactsByEmail(c1.getEmail());
            } catch (Exception e) {
            }
            try {
                service.deleteContactsByEmail(c2.getEmail());
            } catch (Exception e) {
            }

            // Wait a second
            Thread.sleep(delay);

            // Getting a contact must throw an exception
            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not exist in that account.");
            } catch (Exception e) {
            }
            try {
                checkForContactToDisappear(c2.getEmail(), retries, delay);
                fail("Contact2 should not exist in that account.");
            } catch (Exception e) {
            }

            // This call should create both contacts
            boolean reimportUnsubscribedContacts = true;
            SynchronizationReport report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, reimportUnsubscribedContacts, false, false);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }
            try {
                service.getContact(c2.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact2 should now exist but exception occured.");
                throw e;
            }

            assertEquals(report.getCountNewContacts(), 2);
            assertEquals(report.getCountExistingContacts(), 0);
            assertEquals(report.getCountUnsubscribedContacts(), 0);

            // Unsubscribe contact 2
            service.unsubscribeContactsByEmail(c2.getEmail());

            try {
                checkForContactToDisappear(c2.getEmail(), retries, delay);
                fail("Contact2 should not exist in that account.");
            } catch (Exception e) {
            }

            // Now, contact 2 should be recreated
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, reimportUnsubscribedContacts, false, false);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }
            try {
                service.getContact(c2.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact2 should now exist but exception occured.");
                throw e;
            }

            assertEquals(report.getCountNewContacts(), 1);
            assertEquals(report.getCountExistingContacts(), 1);
            assertEquals(report.getCountUnsubscribedContacts(), 0);

            // Unsubscribe contact 2 again
            service.unsubscribeContactsByEmail(c2.getEmail());

            // Wait a second
            Thread.sleep(delay);

            try {
                checkForContactToDisappear(c2.getEmail(), retries, delay);
                fail("Contact2 should not exist in that account.");
            } catch (Exception e) {
            }

            // Now, contact 2 should be ignored
            reimportUnsubscribedContacts = false;
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, reimportUnsubscribedContacts, false, false);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }
            try {
                checkForContactToDisappear(c2.getEmail(), retries, delay);
                fail("Contact2 should not be recreated.");
            } catch (Exception e) {
            }

            assertEquals(report.getCountNewContacts(), 0);
            assertEquals(report.getCountExistingContacts(), 1);
            assertEquals(report.getCountUnsubscribedContacts(), 1);

            // Unsubscribe contact 1, too
            service.unsubscribeContactsByEmail(c1.getEmail());

            // Wait a second
            Thread.sleep(delay);

            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not exist in that account.");
            } catch (Exception e) {
            }

            // Now, nothing should be imported
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, reimportUnsubscribedContacts, false, false);

            // Wait a second
            Thread.sleep(delay);

            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not be recreated.");
            } catch (Exception e) {
            }
            try {
                checkForContactToDisappear(c2.getEmail(), retries, delay);
                fail("Contact2 should not be recreated.");
            } catch (Exception e) {
            }

            assertEquals(report.getCountNewContacts(), 0);
            assertEquals(report.getCountExistingContacts(), 0);
            assertEquals(report.getCountUnsubscribedContacts(), 2);

            // Finally recreate contact 1
            service.createContact(c1, SynchronizationMode.UPDATE);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }

            // Now, contact 1 should be updated
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, reimportUnsubscribedContacts, false, false);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }
            try {
                checkForContactToDisappear(c2.getEmail(), retries, delay);
                fail("Contact2 should not be recreated.");
            } catch (Exception e) {
            }

            assertEquals(report.getCountNewContacts(), 0);
            assertEquals(report.getCountExistingContacts(), 1);
            assertEquals(report.getCountUnsubscribedContacts(), 1);

            // Finally check if previously unsubscribed contact can be changed
            c1.getStandardFields().put(StandardContactField.FIRSTNAME, "Max1Updated");
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, reimportUnsubscribedContacts, false, false);

            // Wait a second
            Thread.sleep(delay);

            try {
                List<StandardContactField> scfs = new ArrayList<>();
                scfs.add(StandardContactField.FIRSTNAME);
                Contact contact = service.getContact(c1.getEmail(), scfs, null);
                assertEquals(contact.getStandardFields().get(StandardContactField.FIRSTNAME), c1.getStandardFields().get(StandardContactField.FIRSTNAME));
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }

            // Clean contacts
            try {
                service.deleteContactsByEmail(c1.getEmail());
            } catch (Exception e) {
            }
            try {
                service.deleteContactsByEmail(c2.getEmail());
            } catch (Exception e) {
            }

        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    private void checkForContactToDisappear(String email, int retries, int delayInMs) throws InterruptedException, MaileonException {
        for (int i = 0; i < retries; i++) {
            // Wait a second
            Thread.sleep(delayInMs);
            service.getContact(email, null, null);
        }
    }

    @Test
    public void testSynchronizeContactsLifecycleWithUpdateOnly() throws Throwable {
        String msg = " - Testing 'testSynchronizeContactsLifecycleWithUpdateOnly'";
        try {

            // Some operations seem to be not complete even if API call returns
            int delay = 1000;
            int retries = 10;

            List<Contact> contacts = new ArrayList<>();

            Contact c1 = new Contact();
            c1.setEmail("maxmustermann1@xqueue.de");
            c1.getStandardFields().put(StandardContactField.FIRSTNAME, "Max1");
            c1.getStandardFields().put(StandardContactField.LASTNAME, "Mustermann1");
            contacts.add(c1);

            // First make sure both are deleted
            try {
                service.deleteContactsByEmail(c1.getEmail());
            } catch (Exception e) {
            }

            // Wait a second
            Thread.sleep(delay);

            // Getting a contact must throw an exception
            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not exist in that account.");
            } catch (Exception e) {
            }

            // Contact1 is not there, should also not be recreated
            boolean updateOnly = true;
            SynchronizationReport report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, true, false, updateOnly);

            assertEquals(report.getCountNewContacts(), 1);
            assertEquals(report.getCountExistingContacts(), 0);
            assertEquals(report.getCountUnsubscribedContacts(), 0);

            // Wait a second
            Thread.sleep(delay);

            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not exist in that account.");
            } catch (Exception e) {
            }

            // This call should now create the contact
            updateOnly = false;
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, true, false, updateOnly);

            assertEquals(report.getCountNewContacts(), 1);
            assertEquals(report.getCountExistingContacts(), 0);
            assertEquals(report.getCountUnsubscribedContacts(), 0);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }

            // Unsubscribe contact 1
            service.unsubscribeContactsByEmail(c1.getEmail());

            // Wait a second
            Thread.sleep(delay);

            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not exist in that account.");
            } catch (Exception e) {
            }

            // Now, contact 1 should be ignored as it is not existing
            updateOnly = true;
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, true, false, updateOnly);

            // When reimporting unsubscribed contacts they count as new
            assertEquals(report.getCountNewContacts(), 1);
            assertEquals(report.getCountExistingContacts(), 0);
            assertEquals(report.getCountUnsubscribedContacts(), 0);

            // Wait a second
            Thread.sleep(delay);

            try {
                checkForContactToDisappear(c1.getEmail(), retries, delay);
                fail("Contact1 should not be recreated.");
            } catch (Exception e) {
            }

            // Now, contact1 should be reimported
            updateOnly = false;
            report = service.synchronizeContacts(contacts, Permission.DOI, SynchronizationMode.UPDATE, false, false, true, false, updateOnly);

            // When reimporting unsubscribed contacts they count as new
            assertEquals(report.getCountNewContacts(), 1);
            assertEquals(report.getCountExistingContacts(), 0);
            assertEquals(report.getCountUnsubscribedContacts(), 0);

            // Wait a second
            Thread.sleep(delay);

            try {
                service.getContact(c1.getEmail(), null, null);
            } catch (Exception e) {
                fail("Contact1 should now exist but exception occured.");
                throw e;
            }

            // Clean up
            service.unsubscribeContactsByEmail(c1.getEmail());

        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Get all contacts that match the filter with the ID DEFAULT_CONTACT_FILTER_ID
     */
    @Test
    public void testGetContactsByFilterIdCompressed() throws Throwable {
        String msg = " - Testing 'retrieve all contact filters matching filter (compressed): " + DEFAULT_CONTACT_FILTER_ID + "'";
        try {
            List<StandardContactField> standardFields = new ArrayList<>();
            standardFields.add(StandardContactField.FIRSTNAME);

            List<String> customFields = new ArrayList<>();

            MaileonContactFiltersService cfs = new MaileonContactFiltersService(TestProperties.getMaileonConfig());
            cfs.setDebug(TestProperties.isDebugStatic());
            cfs.setRequestCompressionEnabled(true);
            cfs.setCompressionEnabled(true);
            Page<ContactFilter> filters = cfs.getContactFilters(10, 1);

            Page<Contact> pContacts = service.getContactsByFilterId(filters.getItems().get(9).getId(), standardFields, customFields, 6, 5);

            if (pContacts == null) {
                if (TestProperties.isReportOnSuccessStatic()) {
                    if (TestProperties.isReportOnSuccessStatic()) {
                        TestProperties.report(msg, true, "No contacts returned (null)", null);
                    }
                }
                return;
            }

            List<Contact> contacts = pContacts.getItems();

            if (contacts != null) {
                StringBuilder sb = new StringBuilder();
                for (Contact contact : contacts) {
                    sb.append(contact.toString()).append("\n");
                }
                if (TestProperties.isReportOnSuccessStatic()) {
                    if (TestProperties.isReportOnSuccessStatic()) {
                        TestProperties.report(msg, true, sb.toString(), null);
                    }
                }
            } else {
                if (TestProperties.isReportOnSuccessStatic()) {
                    if (TestProperties.isReportOnSuccessStatic()) {
                        TestProperties.report(msg, true, "No contacts returned (no list appended)", null);
                    }
                }
            }
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Get all contacts that match the filter with the ID DEFAULT_CONTACT_FILTER_ID
     */
    @Test
    public void testGetContactsByFilterId() throws Throwable {
        String msg = " - Testing 'retrieve all contact filters matching filter: " + DEFAULT_CONTACT_FILTER_ID + "'";
        try {
            List<StandardContactField> standardFields = new ArrayList<>();
            standardFields.add(StandardContactField.FIRSTNAME);

            List<String> customFields = new ArrayList<>();

            MaileonContactFiltersService cfs = new MaileonContactFiltersService(TestProperties.getMaileonConfig());
            cfs.setDebug(TestProperties.isDebugStatic());
            Page<ContactFilter> filters = cfs.getContactFilters(10, 1);

            Page<Contact> pContacts = service.getContactsByFilterId(filters.getItems().get(9).getId(), standardFields, customFields, 6, 5);

            if (pContacts == null) {
                if (TestProperties.isReportOnSuccessStatic()) {
                    if (TestProperties.isReportOnSuccessStatic()) {
                        TestProperties.report(msg, true, "No contacts returned (null)", null);
                    }
                }
                return;
            }

            List<Contact> contacts = pContacts.getItems();

            if (contacts != null) {
                StringBuilder sb = new StringBuilder();
                for (Contact contact : contacts) {
                    sb.append(contact.toString()).append("\n");
                }
                if (TestProperties.isReportOnSuccessStatic()) {
                    if (TestProperties.isReportOnSuccessStatic()) {
                        TestProperties.report(msg, true, sb.toString(), null);
                    }
                }
            } else {
                if (TestProperties.isReportOnSuccessStatic()) {
                    if (TestProperties.isReportOnSuccessStatic()) {
                        TestProperties.report(msg, true, "No contacts returned (no list appended)", null);
                    }
                }
            }
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    /**
     * Get all contacts
     */
    @Test
    public void testGetContacts() throws Throwable {
        String msg = " - Testing 'retrieve all contacts";
        try {
            List<StandardContactField> standardFields = new ArrayList<>();
            standardFields.add(StandardContactField.FIRSTNAME);

            List<String> customFields = new ArrayList<>();

            Page<Contact> pContacts = service.getContacts(standardFields, customFields, 1, 100);

            if (pContacts == null) {
                if (TestProperties.isReportOnSuccessStatic()) {
                    TestProperties.report(msg, true, "No contacts returned (null)", null);
                }
                return;
            }

            List<Contact> contacts = pContacts.getItems();

            if (contacts != null) {
                StringBuilder sb = new StringBuilder();
                for (Contact contact : contacts) {
                    sb.append(contact.toString()).append("\n");
                }
                if (TestProperties.isReportOnSuccessStatic()) {
                    TestProperties.report(msg, true, sb.toString(), null);
                }
            } else {
                if (TestProperties.isReportOnSuccessStatic()) {
                    TestProperties.report(msg, true, "No contacts returned (no list appended)", null);
                }
            }
        } catch (MaileonException e) {
            TestProperties.report(msg, false, null, e);
            throw e;
        }
    }

    @Test
    public void testGetBlockedContacts() throws Throwable {
        try {
            service.getBlockedContacts(new ArrayList<StandardContactField>(), new ArrayList<String>(), 1, 5);
            if (TestProperties.isReportOnSuccessStatic()) {
                TestProperties.report("Got blocked contacts", true, null, null);
            }
        } catch (MaileonException e) {
            TestProperties.report("Failed getting blocked contacts", false, null, e);
            throw e;
        }
    }

    @Test
    public void testCreateCustomField() throws Throwable {
        String fieldName = "test" + new Random().nextInt(10000);
        try {
            service.createCustomField(fieldName, "boolean");
            if (TestProperties.isReportOnSuccessStatic()) {
                TestProperties.report("Create custom field", true, null, null);
            }
        } catch (MaileonException e) {
            TestProperties.report("Failed to create custom field", false, null, e);
            throw e;
        }

        // Clear up
        try {
            service.deleteCustomField(fieldName);
            if (TestProperties.isReportOnSuccessStatic()) {
                TestProperties.report("Create custom field", true, null, null);
            }
        } catch (MaileonException e) {
            TestProperties.report("Failed to create custom field", false, null, e);
            throw e;
        }
    }

    @Test
    public void testCreateCustomFieldFailing() throws Throwable {
        String fieldName = "test" + new Random().nextInt(10000);
        try {
            service.createCustomField(fieldName, "boolean");
            if (TestProperties.isReportOnSuccessStatic()) {
                TestProperties.report("Create custom field", true, null, null);
            }
        } catch (MaileonException e) {
            TestProperties.report("Failed to create custom field", false, null, e);
            throw e;
        }

        // This should fail as field is already available
        try {
            service.createCustomField(fieldName, "boolean");
            if (TestProperties.isReportOnSuccessStatic()) {
                TestProperties.report("Create custom field", true, null, null);
            }
        } catch (MaileonException e) {
            TestProperties.report("Failed to create custom field", false, null, e);
            assertEquals(400, e.getStatusCode());
        }

        // Clear up
        try {
            service.deleteCustomField(fieldName);
            if (TestProperties.isReportOnSuccessStatic()) {
                TestProperties.report("Delete custom field", true, null, null);
            }
        } catch (MaileonException e) {
            TestProperties.report("Failed to create custom field", false, null, e);
            throw e;
        }
    }
}
