package com.maileon.api.mailings.mailing.scheduling;

import com.maileon.api.MaileonBadRequestException;
import com.maileon.api.MaileonException;
import com.maileon.api.MaileonNotFoundException;
import com.maileon.api.mailings.MaileonMailingsService;
import com.maileon.api.mailings.scheduling.MaileonSchedulingService;
import com.maileon.api.mailings.scheduling.Scheduling;
import com.maileon.api.test.TestProperties;
import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.UUID;

import static org.junit.Assert.*;

/**
 * Test driver for the scheduling service
 * 
 * @author Felix Heinrichs | Trusted Mails GmbH | <a href="mailto:felix.heinrichs@trusted-mails.com">felix.heinrichs@trusted-mails.com</a>
 * @author Marcus St&auml;nder | Trusted Mails GmbH | <a href="mailto:marcus.staender@trusted-mails.com">marcus.staender@trusted-mails.com</a>
 *
 */
public class MaileonSchedulingTestDriver {

    // For our current development setup

    private static final String DEFAULT_DATE_PAST = "2013-08-26";
    private static final String DEFAULT_DATE = "2020-08-26";
    private static final String DEFAULT_HOURS = "14";
    private static final String DEFAULT_MINUTES = "0";

    private static final String DEFAULT_DATE2 = "2020-06-29";
    private static final String DEFAULT_HOURS2 = "1";
    private static final String DEFAULT_MINUTES2 = "30";


    public static Long mailingId;



    protected static MaileonMailingsService service = null;
    protected static MaileonSchedulingService schedulingService = null;

    private static TestProperties testProperties;

    @BeforeClass
    public static void initialize() {
        testProperties = new TestProperties();
        service = new MaileonMailingsService(testProperties.getMaileonConfiguration());
        service.setDebug(testProperties.isDebug());

        schedulingService = new MaileonSchedulingService(testProperties.getMaileonConfiguration());
        schedulingService.setDebug(testProperties.isDebug());
    }

    @Before
    public void initializeTest() {
        mailingId = null;
    }

    @After
    public void cleanTest() {
        service.setDebug(false);
        if (mailingId != null) {
            try {
                service.deleteMailing(mailingId);
            } catch (MaileonException e) {
                e.printStackTrace();
            }
        }
        service.setDebug(testProperties.isDebug());
    }


    /**
     * Check if a schedule for the mailing with the given mailingId exists and return details
     */
    @Test(expected = MaileonNotFoundException.class)
    public void testRetrieveFailing() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        schedulingService.getSchedule(mailingId);
    }

    @Test(expected = MaileonNotFoundException.class)
    public void testUpdateFailing() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        schedulingService.updateSchedule(mailingId, DEFAULT_DATE2, DEFAULT_HOURS2, DEFAULT_MINUTES2);
    }

    @Test(expected = MaileonBadRequestException.class)
    public void testCreateScheduleInPast() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        schedulingService.updateSchedule(mailingId, DEFAULT_DATE_PAST, DEFAULT_HOURS, DEFAULT_MINUTES);
    }

    @Test(expected = MaileonBadRequestException.class)
    public void testCreateScheduleForTrigger() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        schedulingService.updateSchedule(mailingId, DEFAULT_DATE, DEFAULT_HOURS, DEFAULT_MINUTES);
    }

    /**
     * Check if creating a schedule for the mailing with the given mailingId works
     */
    @Test
    public void testCreate() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        String apiReportedType = service.getType(mailingId);
        assertEquals(MaileonMailingsService.MailingType.REGULAR.getApiString(), apiReportedType);

        schedulingService.setSchedule(mailingId, DEFAULT_DATE, DEFAULT_HOURS, DEFAULT_MINUTES);

        Scheduling schedule = schedulingService.getSchedule(mailingId);

        assertEquals(DEFAULT_DATE, schedule.getDate());
        assertEquals(DEFAULT_HOURS, schedule.getHours());
        assertEquals(DEFAULT_MINUTES, schedule.getMinutes());
    }


    /**
     * Check if updating a schedule for the mailing with the given mailingId works
     */
    @Test
    public void testUpdate() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        String apiReportedType = service.getType(mailingId);
        assertEquals(MaileonMailingsService.MailingType.REGULAR.getApiString(), apiReportedType);

        schedulingService.setSchedule(mailingId, DEFAULT_DATE, DEFAULT_HOURS, DEFAULT_MINUTES);

        Scheduling schedule = schedulingService.getSchedule(mailingId);

        assertEquals(DEFAULT_DATE, schedule.getDate());
        assertEquals(DEFAULT_HOURS, schedule.getHours());
        assertEquals(DEFAULT_MINUTES, schedule.getMinutes());

        schedulingService.updateSchedule(mailingId, DEFAULT_DATE2, DEFAULT_HOURS2, DEFAULT_MINUTES2);

        schedule = schedulingService.getSchedule(mailingId);

        assertEquals(DEFAULT_DATE2, schedule.getDate());
        assertEquals(DEFAULT_HOURS2, schedule.getHours());
        assertEquals(DEFAULT_MINUTES2, schedule.getMinutes());
    }


    /**
     * Check if updating a schedule for the mailing with the given mailingId works
     */
    @Test(expected = MaileonBadRequestException.class)
    public void testSetDuplicate() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        schedulingService.setSchedule(mailingId, DEFAULT_DATE, DEFAULT_HOURS, DEFAULT_MINUTES);
        schedulingService.setSchedule(mailingId, DEFAULT_DATE2, DEFAULT_HOURS2, DEFAULT_MINUTES2);
    }


    /**
     * Check if deleting a schedule for the mailing with the given mailingId exists and return details
     */
    @Test public void testDelete() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.REGULAR);
        assertNotNull(mailingId);

        String apiReportedType = service.getType(mailingId);
        assertEquals(MaileonMailingsService.MailingType.REGULAR.getApiString(), apiReportedType);

        schedulingService.setSchedule(mailingId, DEFAULT_DATE, DEFAULT_HOURS, DEFAULT_MINUTES);

        Scheduling schedule = schedulingService.getSchedule(mailingId);

        assertEquals(DEFAULT_DATE, schedule.getDate());
        assertEquals(DEFAULT_HOURS, schedule.getHours());
        assertEquals(DEFAULT_MINUTES, schedule.getMinutes());

        schedulingService.deleteSchedule(mailingId);

        try {

            schedulingService.getSchedule(mailingId);
            fail("Schedule should be deleted.");
        } catch (MaileonNotFoundException e) {

        }

    }


    @Test public void testScheduleFromDate() throws Exception {

        String testDate = "29.12.2016 14:00:14";
        DateFormat formatter = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss");
        Date date = formatter.parse(testDate);

        Scheduling scheduling = new Scheduling(date);

        assertEquals("2016-12-29", scheduling.getDate());
        assertEquals("14", scheduling.getHours());
        assertEquals("00", scheduling.getMinutes());

    }

    private String getRandomName() {
        return "API_"+ UUID.randomUUID();
    }

    private String getRandomSubject() {
        return "Subject_"+ UUID.randomUUID();
    }
}
