package com.maileon.api.mailings.mailing.scheduling;

import com.maileon.api.MaileonBadRequestException;
import com.maileon.api.MaileonException;
import com.maileon.api.mailings.MaileonMailingsService;
import com.maileon.api.mailings.trigger.DispatchOptions;
import com.maileon.api.mailings.trigger.DispatchOptionsXmlSerializer;
import com.maileon.api.test.TestProperties;
import com.maileon.api.xml.XmlComparator;
import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import java.util.UUID;

import static org.junit.Assert.*;

/**
 * Test driver for creation and management of trigger mailings
 * 
 * @author mstaender
 *
 */
public class TriggerMailingTest {

    public static Long mailingId;

    protected static MaileonMailingsService service = null;
    private static TestProperties testProperties;

    @BeforeClass
    public static void initialize() {
        testProperties = new TestProperties();
        service = new MaileonMailingsService(testProperties.getMaileonConfiguration());
        service.setDebug(testProperties.isDebug()); // Enable this for more debugging information
    }

    @Before
    public void initializeTest() {
        mailingId = null;
    }

    @After
    public void cleanTest() {
        service.setDebug(false);
        if (mailingId != null) {
            try {
                service.deleteMailing(mailingId);

            } catch (Exception e) {
//                e.printStackTrace();
                try {
                    service.deactivateTriggerMailing(mailingId);
                } catch (Exception ex) {
                }
            }
        }
        service.setDebug(testProperties.isDebug());
    }

    @Test
    public void testCreateDuplicateTrigger() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        // Should fail
        try {
            service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
            fail("Bad request expected due to duplicate mail");
        } catch (MaileonBadRequestException e) {
            assertTrue(e.getMessage().contains("a mailing exists with the name : '"+name+"'"));
        }
    }

    @Test
    public void testGetDeletedTrigger() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        String apiReportedSubject = service.getSubject(mailingId);
        assertEquals(subject, apiReportedSubject);

        service.deleteMailing(mailingId);

        String responseSubject = service.getSubject(mailingId);
        assertNull(responseSubject);
    }

    @Test
    public void testCreateTrigger() throws Exception {
        try {
            String subject = getRandomSubject();
            String name = getRandomName();

            mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
            assertNotNull(mailingId);

            String apiReportedSubject = service.getSubject(mailingId);
            assertEquals(subject, apiReportedSubject);

            String apiReportedType = service.getType(mailingId);
            assertEquals(MaileonMailingsService.MailingType.TRIGGER.getApiString(), apiReportedType);

            String apiReportedName = service.getName(mailingId);
            assertEquals(name, apiReportedName);
        } catch (MaileonException e) {
            fail(e.getMessage());
            throw e;
        }
    }

    @Test
    public void testCreateAndStartHtmlTrigger() throws Exception {
        try {
//            String html = FileUtils.readFileToString(new File(this.getClass().getClassLoader().getResource("medien-jaeger-html-test.html").getFile()), StandardCharsets.UTF_8);
            String html = "<html></html>";

            String subject = getRandomSubject();
            String name = getRandomName();

            mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
            assertNotNull(mailingId);

            service.setHtmlContent(mailingId, html);

            service.setSenderAddress(mailingId, "test");

            DispatchOptions options = DispatchOptionsXmlSerializer.deserialize("<?xml version=\"1.0\"?><dispatch_options><type>MULTI</type><target>CONTACTFILTER</target><contact_filter_id>2765</contact_filter_id><interval>WEEK</interval><day_of_week>7</day_of_week><hours>12</hours><minutes>30</minutes><start_trigger>true</start_trigger></dispatch_options>");
            service.setTriggerDispatchOptions(mailingId, options);

            Thread.sleep(1000);

            service.deactivateTriggerMailing(mailingId);

        } catch (MaileonException e) {
            e.printStackTrace();
            fail(e.getMessage());
            throw e;
        }
    }

    @Test
    public void testCreateDoi() throws Exception {
        try {
            String subject = getRandomSubject();
            String name = getRandomName();

            mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.DOI);
            assertNotNull(mailingId);

            String apiReportedSubject = service.getSubject(mailingId);
            assertEquals(subject, apiReportedSubject);

            String apiReportedType = service.getType(mailingId);
            assertEquals(MaileonMailingsService.MailingType.DOI.getApiString(), apiReportedType);

            String apiReportedName = service.getName(mailingId);
            assertEquals(name, apiReportedName);
        } catch (MaileonException e) {
            fail(e.getMessage());
            throw e;
        }
    }

    @Test
    public void testSetTriggerDispatchLogicNonXsdOrder() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        String schedule = "<?xml version=\"1.0\"?><dispatch_options><event>58</event><type>SINGLE</type><speed_level>low</speed_level><start_trigger>false</start_trigger></dispatch_options>";
        service.setTriggerDispatchOptions(mailingId, schedule);

        String expectedResponse =
                "<dispatch_options>" +
                "<type>SINGLE</type>" +
                "<event>58</event>" +
                "<event_id>58</event_id>" +
                "<start_trigger>false</start_trigger>" +
                "<target>EVENT</target>" +
                "</dispatch_options>";
        String triggerDispatchSchedule = service.getTriggerDispatchScheduleRaw(mailingId);

        XmlComparator.assertSimilar(expectedResponse, triggerDispatchSchedule);
    }

    @Test(expected = MaileonBadRequestException.class)
    public void testSetTriggerDispatchLogicNonValidXsdScheme() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        String schedule = "<?xml version=\"1.0\"?><dispatch_options><eventX>58</eventX><type>SINGLE</type><speed_level>low</speed_level><start_trigger>false</start_trigger></dispatch_options>";
        service.setTriggerDispatchOptions(mailingId, schedule);
    }

    @Test
    public void testSetTriggerDispatchLogicInXsdOrderSingle() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        String schedule = "<?xml version=\"1.0\"?><dispatch_options><type>SINGLE</type><event>58</event><speed_level>low</speed_level><start_trigger>false</start_trigger></dispatch_options>";
        service.setTriggerDispatchOptions(mailingId, schedule);

        String expectedResponse =
                "<dispatch_options>" +
                "<type>SINGLE</type>" +
                "<event>58</event>" +
                "<event_id>58</event_id>" +
                "<start_trigger>false</start_trigger>" +
                "<target>EVENT</target>" +
                "</dispatch_options>";
        String triggerDispatchSchedule = service.getTriggerDispatchScheduleRaw(mailingId);

        XmlComparator.assertSimilar(expectedResponse, triggerDispatchSchedule);
    }

    @Test
    public void testSetTriggerDispatchLogicInXsdOrderMultiEventHour() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        Thread.sleep(3000);

        String schedule = "<?xml version=\"1.0\"?><dispatch_options><type>MULTI</type><target>EVENT</target><event>58</event><interval>HOUR</interval><minutes>30</minutes></dispatch_options>";
        service.setTriggerDispatchOptions(mailingId, schedule);

        String expectedResponse =
                "<dispatch_options>" +
                "<type>MULTI</type>" +
                "<event>58</event>" +
                "<event_id>58</event_id>" +
                "<start_trigger>false</start_trigger>" +
                "<target>EVENT</target>" +
                "<interval>HOUR</interval>" +
                "<day_of_month>0</day_of_month>" +
                "<hours>0</hours>" +
                "<minutes>30</minutes>" +
                "<day_of_week>0</day_of_week>" +
                "</dispatch_options>";
        String triggerDispatchSchedule = service.getTriggerDispatchScheduleRaw(mailingId);

        XmlComparator.assertSimilar(expectedResponse, triggerDispatchSchedule);
    }

    @Test
    public void testSetTriggerDispatchLogicInXsdOrderMultiContactfilterMonthly() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        String schedule = "<?xml version=\"1.0\"?><dispatch_options><type>MULTI</type><target>CONTACTFILTER</target><contact_filter_id>2765</contact_filter_id><interval>MONTH</interval><day_of_month>15</day_of_month><hours>12</hours><minutes>30</minutes></dispatch_options>";
        service.setTriggerDispatchOptions(mailingId, schedule);

        String expectedResponse =
                "<dispatch_options>" +
                "<type>MULTI</type>" +
                "<start_trigger>false</start_trigger>" +
                "<target>CONTACTFILTER</target>" +
                "<interval>MONTH</interval>" +
                "<day_of_month>15</day_of_month>" +
                "<hours>12</hours>" +
                "<minutes>30</minutes>" +
                "<day_of_week>0</day_of_week>" +
                "<contact_filter_id>2765</contact_filter_id>" +
                "</dispatch_options>";
        String triggerDispatchSchedule = service.getTriggerDispatchScheduleRaw(mailingId);

        XmlComparator.assertSimilar(expectedResponse, triggerDispatchSchedule);
    }

    @Test
    public void testSetTriggerDispatchLogicInXsdOrderMultiContactfilterWeekly() throws Exception {
        String subject = getRandomSubject();
        String name = getRandomName();

        mailingId = service.createMailing(name, subject, MaileonMailingsService.MailingType.TRIGGER);
        assertNotNull(mailingId);

        DispatchOptions options = DispatchOptionsXmlSerializer.deserialize("<?xml version=\"1.0\"?><dispatch_options><type>MULTI</type><target>CONTACTFILTER</target><contact_filter_id>2765</contact_filter_id><interval>WEEK</interval><day_of_week>1</day_of_week><hours>12</hours><minutes>30</minutes></dispatch_options>");
        service.setTriggerDispatchOptions(mailingId, options);

        String expectedResponse = "<dispatch_options>" +
                "<type>MULTI</type>" +
                "<start_trigger>false</start_trigger>" +
                "<target>CONTACTFILTER</target>" +
                "<interval>WEEK</interval>" +
                "<day_of_month>0</day_of_month>" +
                "<hours>12</hours>" +
                "<minutes>30</minutes>" +
                "<day_of_week>1</day_of_week>" +
                "<contact_filter_id>2765</contact_filter_id>" +
                "</dispatch_options>";
        String triggerDispatchSchedule = service.getTriggerDispatchScheduleRaw(mailingId);

        // Test deserialization
        DispatchOptions dispatchOptions = service.getTriggerDispatchSchedule(mailingId);

        assertEquals("MULTI", dispatchOptions.getType());
        assertFalse(dispatchOptions.isStartTrigger());
        assertEquals("CONTACTFILTER", dispatchOptions.getTarget());
        assertEquals("WEEK", dispatchOptions.getInterval());
        assertEquals(new Integer(0), dispatchOptions.getDayOfMonth());
        assertEquals(new Integer(12), dispatchOptions.getHours());
        assertEquals(new Integer(30), dispatchOptions.getMinutes());
        assertEquals(new Integer(1), dispatchOptions.getDayOfWeek());
        assertEquals(new Long(2765), dispatchOptions.getContactFilterId());

        XmlComparator.assertSimilar(expectedResponse, triggerDispatchSchedule);
    }

    @Test
    public void testTriggerDispatchSerialization() throws Exception {

        DispatchOptions options = DispatchOptionsXmlSerializer.deserialize("<?xml version=\"1.0\"?><dispatch_options><type>MULTI</type><target>CONTACTFILTER</target><contact_filter_id>2765</contact_filter_id><interval>WEEK</interval><day_of_week>1</day_of_week><hours>12</hours><minutes>30</minutes></dispatch_options>");

        assertEquals("MULTI", options.getType());
        assertFalse(options.isStartTrigger());
        assertEquals("CONTACTFILTER", options.getTarget());
        assertEquals("WEEK", options.getInterval());
        assertNull(options.getDayOfMonth());
        assertEquals(new Integer(12), options.getHours());
        assertEquals(new Integer(30), options.getMinutes());
        assertEquals(new Integer(1), options.getDayOfWeek());
        assertEquals(new Long(2765), options.getContactFilterId());
    }

    @Test
    public void testTriggerDispatchDeserialization() throws Exception {

        DispatchOptions options = new DispatchOptions();
        options.setType("MULTI");
        options.setStartTrigger(true);
        options.setTarget("CONTACTFILTER");
        options.setInterval("WEEK");
        options.setDayOfMonth(0);
        options.setHours(12);
        options.setMinutes(30);
        options.setDayOfWeek(1);
        options.setContactFilterId(2765);

        String expectedResponse = "<dispatch_options>" +
                "<type>MULTI</type>" +
                "<start_trigger>true</start_trigger>" +
                "<target>CONTACTFILTER</target>" +
                "<interval>WEEK</interval>" +
                "<day_of_month>0</day_of_month>" +
                "<hours>12</hours>" +
                "<minutes>30</minutes>" +
                "<day_of_week>1</day_of_week>" +
                "<contact_filter_id>2765</contact_filter_id>" +
                "</dispatch_options>";

        XmlComparator.assertSimilar(expectedResponse, DispatchOptionsXmlSerializer.serialize(options).asXML());
    }

    private String getRandomName() {
        return "API_"+ UUID.randomUUID();
    }

    private String getRandomSubject() {
        return "Subject_"+ UUID.randomUUID();
    }
}
